/***********************************************************************
frogd, daemon pour l'utilisation des capteurs meteo de http://www.froggyhome.com

Copyright (C) 2001 (Regis Barbier)
Ce programme est libre, vous pouvez le redistribuer et/ou le modifier selon 
les termes de la Licence Publique Gnrale GNU publie par la Free Software 
Foundation (version 2 ou bien toute autre version ultrieure choisie par vous).

Ce programme est distribu car potentiellement utile, mais SANS AUCUNE 
GARANTIE, ni explicite ni implicite, y compris les garanties de 
commercialisation ou d'adaptation dans un but spcifique. Reportez-vous  la 
Licence Publique Gnrale GNU pour plus de dtails.

Vous devez avoir reu une copie de la Licence Publique Gnrale GNU en mme 
temps que ce programme. Si ce n'est pas le cas, crivez  la Free Software 
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, tats-Unis.
*************************************************************************/

#include "frogd.h"
#include "frogd_comm.h"

//---------------------------------------------------------------
// Calculate the checksum of a frame
//---------------------------------------------------------------
int CalcCheckSum(Frog *frog, unsigned char* input,  int framesize, char* checksum) {
   unsigned  char tempo[FROG_MAX_FRAME_SIZE];
   unsigned char check;
   int i;
    
   Decode(input+1, tempo, framesize-1);
   
   check='G';
   for (i=0; i<(framesize-1)/2; i++) {
      check=check^tempo[i];
   }
 
   sprintf( checksum, "%.2X", check); 
   if (frog->debug) printf("Checksum is %s, framesize is %d\n", checksum, framesize);
   return 1;

}

//---------------------------------------------------------------
// decode the input string : translate from the 2 char format to the 1 byte
//---------------------------------------------------------------
int Decode(unsigned char* input,  unsigned char* output, int size) {
   unsigned  int  a;
   int i;

   for(i=0; i<(size/2) ;i++) {
      sscanf( input+(i*2), "%2x", &a);
      *(output+i) =  a; 
//      write(1, input+(i*2), 2);
//      printf("dec - %.2X \n",*(output+i) );
   }
   return 1;
}
 
//---------------------------------------------------------------
// Open the serial port put the file descriptor in frog->fd
//---------------------------------------------------------------
int OpenPort(Frog* frog) {
   if ( (frog->fd = open(frog->device, O_RDWR  )) <= 0) {
      perror("Can't open serial port");
      return -1;
   }

   if (!isatty(frog->fd)) {
      printf ("device is not a tty device\n");
      close(frog->fd);
      return -1;
  }

   flock(frog->fd, LOCK_EX );
   if (frog->debug)  printf("Serial port Opened\n");
   return 1;
}

//---------------------------------------------------------------
// Close the serial port
//---------------------------------------------------------------
int ClosePort(Frog* frog) {
   if ( (frog->fd = close(frog->fd)) < 0) {
      perror("Can't close serial port");
      return -1;
   }

   if (frog->debug)  printf("Serial port closed\n");
   return 1;
}

//---------------------------------------------------------------
// Configure the serial port where the frog is attached
//---------------------------------------------------------------
int ConfigurePort (Frog* frog) {
   struct termios attr;

   bzero(&attr, sizeof(attr)); 
   attr.c_cflag |=  (CS8 | CLOCAL | CREAD);
   if (frog->stopbit == 2) {
      attr.c_cflag |=  CSTOPB;
   }
   attr.c_iflag = BRKINT | INPCK;
   attr.c_oflag =  0;
   attr.c_lflag  = 0;
 
   attr.c_cc[VTIME]    = FROG_READ_DELAY;
   attr.c_cc[VMIN]     = 0;

   if (cfsetispeed(&attr, frog->baud)== -1) {
      perror("Can't set the serial port input speed");
      return -1;
   }
   if (cfsetospeed(&attr, frog->baud)== -1) {
      perror("Can't set the serial port output speed");
      return -1;
   }      

   if (tcsetattr(frog->fd,TCSANOW,&attr)== -1) { 
      perror("Can't set the serial port configuration");
      return -1;
   }
   
   if (frog->debug) printf("Setattr (ConfigurePort) done\n");
   return(1);
}

//---------------------------------------------------------------
// Reset the frog by switching it OFF and ON 
//---------------------------------------------------------------
int ResetFrog(Frog* frog) {
   int status;
   if (ioctl(frog->fd, TIOCMGET, &status) == -1 ) 
      perror("Can't get the serial port modem line status");
   
   status |=  TIOCM_DTR;
   status &= ~TIOCM_RTS;
   if (ioctl(frog->fd, TIOCMSET, &status) == -1 ) {
      perror("Can't set the serial port modem line status");
      return -1;
   }
   sleep(FROG_RESET_DELAY);
   
   status &=  ~TIOCM_DTR ;
   status |=  TIOCM_RTS ;

   if (ioctl(frog->fd, TIOCMSET, &status) == -1 ) {
      perror("Can't set the serial port modem line status");
      return -1;
   }
   
   sleep(FROG_RESET_DELAY);
   tcflush(frog->fd,TCIOFLUSH );
   
   if (frog->debug) printf("Frog reset done\n");
   return(1);
}

//---------------------------------------------------------------
// Send a frame to the frog and wait for the response.
//---------------------------------------------------------------
int AskFrog(char* command, Frog* frog, char* response) {
   unsigned char frame[FROG_MAX_FRAME_SIZE];
   unsigned char checksum[FROG_MAX_CHECKSUM_SIZE];
   int gotend=FALSE;
   int i;

   sprintf(frame, "G%s", command);   
   CalcCheckSum(frog, frame, strlen(frame), checksum);
   sprintf(frame, "G%s%sZ", command, checksum);
   
   if (write(frog->fd, frame , strlen(frame) ) != strlen(frame)) {
      printf ("Can't write to the serial port");
      return -1;
   }

   if (frog->debug) printf("frame sent : %s\n", frame);
   
   i = 0;
   while (!gotend) {
      if (read(frog->fd, response+i, 1)  <=0) {
         //perror("read AskFrog:");
         return -1;
      }
      if (response[i] == END_FRAME) gotend = TRUE;
      i++;
   }

   CalcCheckSum(frog, response, i-3, checksum);
   if (strncmp(response+i-3, checksum, 2)) {
      printf ("Bad checksum\n");
      return -1;
   }
   if (frog->debug) printf("the checksum is %s\n", checksum);
   return i;
}


//---------------------------------------------------------------
// ask the frog to return its identification.
// firmware, serial number, number of sensors and type ...
//---------------------------------------------------------------
int GetIdentification(Frog* frog) {
   unsigned char tab[255];
   unsigned char* reponse;
   int nbread; 
   char* com="00";            // command "00" => get frog identification
   int i;
   int offset;
 
   if (frog->debug) printf("We are in GetIdentification\n");
   reponse = tab;
   
   if ( (nbread = AskFrog(com, frog, reponse)) < 0) {
      printf("Askfrog not successfull \n");
      return -1;
   }
  
   reponse++;      // strip leading G

   if (frog->debug) {
      for (i=0; i<nbread/2; i++) {
         printf("offset %d - value %c%c \n", i*2, reponse[i*2], reponse[i*2+1]); 
      }
      printf("frame size : %d octet\n", nbread );
   }

   sscanf(reponse+2, "%2x",  &(frog->firmware));
   sscanf(reponse+4, "%2x",  &(frog->partnumber));
   sscanf(reponse+6, "%8x",  &(frog->serialnumber));
   sscanf(reponse+14, "%2x", &(frog->nbsensor));
   
   offset = 16;
   for (i=0; i<frog->nbsensor; i++) {
      sscanf(reponse+ offset , "%2x", &(frog->sensor[i].identification));
      sscanf(reponse+ offset +2 , "%2x", &(frog->sensor[i].phymesurementsize));
      sscanf(reponse+ offset +4 , "%2x", &(frog->sensor[i].datasize));
      Decode( reponse+ offset + 6, (unsigned char*) &(frog->sensor[i].data), frog->sensor[i].datasize*2 );
      offset += 6 + frog->sensor[i].datasize*2;
   }
   
   return 1;
}











