import asyncore, asynchat, socket
import os, sys, time

VERSION = 'v0.1'
PORT = 8881
DATA_ROOT='/home/dsoulayr/public_html/frogd'

class User(object):
    def __init__(self, aMachine, aName = ''):
        self.ip, self.port = aMachine
        self.name = aName

    def __call__(self):
        return self.name + '(' + self.ip + ':' + str(self.port) + ')'

class MainServer(asyncore.dispatcher):
    def __init__(self, aPort):
        print 'Initializing main server on port', aPort, '...',
        asyncore.dispatcher.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.bind(('', aPort))
        self.listen(5)
        print 'Ok.'

    def handle_accept(self):
        aSock, aAddress = self.accept()
        print 'Connected from', aAddress
        FrogServer(aSock)


class FrogServer(asynchat.async_chat):
    def __init__(self, *aArgs):
        print 'Initializing Frog individual server...',
        asynchat.async_chat.__init__(self, *aArgs)
        self.set_terminator('\n')
        self.data = ''
        self.user = User(self.getpeername())
        self.file = None
        print 'Ok.'


    def authenticate(self, aUser):
        print aUser, 'connected on', self.user()
        self.user.name = aUser
        self.file = open(os.path.join(DATA_ROOT, self.user.name), 'a')

    def storeData(self, aFrogData):
        """Write the data to a file.

        Format is dd/MM/yyyy:hh/mm:temp:pres:hygro:[comment]
        """
        if self.user.name == '':
            print 'Data received without authentification.', \
                  'Please connect first.'
            return
        lData, lComment = aFrogData.split(':')
        lTemperature, lPressure, lHygroscopy = lData.split(' ')
        print '%sdeg, %smBar, %s%% from %s' % \
              (lTemperature, lPressure, lHygroscopy, self.user())
        
        lLine = self._formatDate() + ':' \
                + lTemperature + ':' + lPressure + ':' + lHygroscopy \
                + ':' + lComment.replace(' ', '%20') + os.linesep
        
        self.file.write(lLine)
        self.file.flush()
        
    def collect_incoming_data(self, aData):
        print 'Received %d bytes from %s' % (len(aData), self.user())
        self.data = ''.join(aData)

    def found_terminator(self):
        if self.data.startswith('user'):
            self.authenticate(self.data[4:].strip())
        elif self.data.startswith('data'):
            self.storeData(self.data[4:].strip())
        else:
            print 'Unknown command from', self.user
        self.data = []

    def handle_close(self):
        print self.user(),'disconnected'
        self.close()
        self.file.close()

    def _formatDate(self):
        lTimeTuple = time.localtime(time.time())     
        return time.strftime('%d/%m/%Y:%H/%M', lTimeTuple)

print '## Frog Python Daemon.', VERSION
try:
    Server = MainServer(PORT)
except socket.error, (aCode, aMessage):
    print 'Error:', aMessage
    sys.exit()

try:
    asyncore.loop()
except KeyboardInterrupt:
    Server.close()
