"""

"""

import asyncore, asynchat, socket
import os, sys, zipfile
import time, datetime
import getopt


VERSION = 'v0.2'
DEFAULT_PORT = 8882
DATA_ROOT='/home/dsoulayr/public_html/frogd'

# Textual data
NOTICE = """NetFrogd version 2.0, Copyright (C) 2003 David Soulayrol
NetFrogd comes with ABSOLUTELY NO WARRANTY; for details type `NetFrogd -w'.
This is free software, and you are welcome to redistribute it under
certain conditions; see the COPYING file.
"""

WARRANTY = """  NO WARRANTY

  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS
TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
REPAIR OR CORRECTION.

  12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,
INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING
OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED
TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY
YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER
PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGES.
"""

class User(object):
    """A wrapper for user information.

    A user is actually a machine running the NetFrog client. We
    currently only know its IP address, port, and hostname.
    """

    def __init__(self, aMachine, aName = ''):
        """The unique constructor of a user.
        """
        self.ip, self.port = aMachine
        self.name = aName

    def __call__(self):
        """Returns a fully informative string for this user.
        """
        return self.name + '(' + self.ip + ':' + str(self.port) + ')'


class LogLevels(object):
    def __init__(self):
        self.isQuiet = False
        self.isVerbose = False
        self.isDebug = False

        
class Logger(object):
    INFO = '##'
    MORE = '  '
    ERROR = '!!'
    DEBUG = '++'
    
    def __init__(self, aLogLevels):
        self.__levels = aLogLevels
        
    def debug(self, aMessage):
        if self.__levels.isDebug:
            print self.DEBUG, self.__processMessage(aMessage)

    def inform(self, aMessage):
        if not self.__levels.isQuiet:
            print self.INFO, self.__processMessage(aMessage)

    def log(self, aMessage):
        if self.__levels.isVerbose or self.__levels.isDebug:
            print self.MORE, self.__processMessage(aMessage)

    def warn(self, aMessage):
        print self.ERROR, self.__processMessage(aMessage)

    def __processMessage(self, aMessage):
        if not aMessage.endswith('.'):
            aMessage += '.'
        return aMessage


class Translator(object):
    DATE = 0
    TIME = 1
    PRESSURE = 2
    HYGROSCOPY = 3
    TEMPERATURE = 4
    
    def __init__(self, aLogger):
        self.__logger = aLogger

    def translate(self, aUser, aFile):
        self.__logger.inform('Translating legacy data file ' + aFile)
        lPath, lFilename = os.path.split(aFile)
        os.chdir(lPath)
        if not os.path.isfile(lFilename):
            self.__logger.warn(lFilename + ' is not a file.')
            return
        
        lInFile = open(lFilename, 'r')
        lOutFile = None
        lLine = None
        lFutureDate = datetime.datetime(1970, 1, 1)
        while lLine != '':
            lDate, lLine = self.__translateLine(lInFile.readline())
            if lLine != '':
                if lDate > lFutureDate:
                    lCurrentDate, lFutureDate = self.__getWeekDatesFor(lDate)
                    lOutFile = self.__openOutFile(lOutFile, aUser, lCurrentDate)
                lOutFile.write(lLine)

        lInFile.close()
        lOutFile.close()

    def __translateLine(self, aLine):
        if aLine == '':
            return None, ''

        lSplitLine = aLine.split()
        lDate = self.__getDateFor(lSplitLine)
        lNewLine = lSplitLine[self.DATE] + ':' + \
                   lSplitLine[self.TIME].replace(':', '/') + ':' + \
                   lSplitLine[self.TEMPERATURE] + ':' + \
                   lSplitLine[self.PRESSURE] + ':' + \
                   lSplitLine[self.HYGROSCOPY] + os.linesep
        return lDate, lNewLine

    def __openOutFile(self, aFile, aUser, aDate):
        """Opens or reopens a new result file for the translation.
        """
        if aFile != None and not aFile.closed:
            aFile.close()
        lOutFilename = aUser + '-' + time.strftime('%Y%m%d', aDate.timetuple())
        self.__logger.log('creating new file ' + lOutFilename)
        return open(lOutFilename, 'w')
        
    def __getDateFor(self, aSplitLine):
        lDay, lMonth, lYear = aSplitLine[self.DATE].split('/')
        return datetime.datetime(int(lYear), int(lMonth), int(lDay))

    def __getWeekDatesFor(self, aDate):
        """Returns the date of the beginning of the concerned week.
        """
        lTuple = aDate.timetuple()
        if lTuple.tm_wday == 6:
            lNextDelta = datetime.timedelta(days = 7)
            return aDate, aDate + lNextDelta
        else:
            lPrevDelta = datetime.timedelta(days = - (lTuple.tm_wday + 2))
            lNextDelta = datetime.timedelta(days = 7 - (lTuple.tm_wday + 2))
            return aDate + lPrevDelta, aDate + lNextDelta


class Dispatcher(object):
    def __init__(self):
        self.__port = DEFAULT_PORT
        self.__server = None
        self.__logger = None

        self.__export = None
        self.__user = None
        
        lLogLevels = self.__readCommandLine(sys.argv[1:])
        if lLogLevels == None:
            sys.exit()
        self.__logger = Logger(lLogLevels)        

    def printUsage(self):
        """Prints a short notice about NetFrogd usage.
        """
        print NOTICE
        print "Usage: python netfrogd.py [options]\n"

    def printGplWarranty(self):
        """Prints the warranty section of the GPL license.
        """
        print WARRANTY

    def process(self):
        if self.__export != None:
            self.__processTranslation()
        else:
            self.__spawnMainServer()
        
    def interrupt(self):
        if self.__server != None:
            self.__server.close()

    def __processTranslation(self):
        if self.__user != None:
            lTranslator = Translator(self.__logger)
            lTranslator.translate(self.__user, self.__export)
        else:
            self.__logger.warn('A user name must be specified.')

    def __spawnMainServer(self):
        try:
            if self.__server == None:
                self.__server = MainServer(self.__port, self.__logger)
                asyncore.loop()

        except socket.error, (aCode, aMessage):
            self.__logger.warn('__spawnMainServer error: ' + aMessage)
            sys.exit()

    def __readCommandLine(self, aLine):
        # Command line options
        OPT_WARRANTY = 'warranty'
        OPT_DEBUG = 'debug'
        OPT_VERBOSE = 'verbose'
        OPT_QUIET = 'quiet'

        OPT_TRANSLATE = 'translate='
        OPT_USER = 'user='
        
        OPT_LIST = 'wdvqt:u:'
        LONG_OPT_LIST = [OPT_WARRANTY, OPT_DEBUG, OPT_VERBOSE, OPT_QUIET]

        try:
            lOptlist, lArgs = getopt.getopt(aLine, OPT_LIST, LONG_OPT_LIST)
        except getopt.GetoptError:
            self.printUsage()
            return None

        if len(lArgs) > 0:
            self.printUsage()
            return None

        lLogLevels = LogLevels()
        for lOptName, lOptArg in lOptlist:
            if lOptName in ('-w', '--' + OPT_WARRANTY):
                printGplWarranty()
                return None
            if lOptName in ('-d', '--' + OPT_DEBUG):
                if not lLogLevels.isQuiet:
                    lLogLevels.isDebug = True
                else:
                    pass #print ERROR, 'debug option conflicts with quiet.'
            if lOptName in ('-v', '--' + OPT_VERBOSE):
                if not lLogLevels.isQuiet:
                    lLogLevels.isVerbose = True
                else:
                    pass #print ERROR, 'verbose option conflicts with quiet.'
            if lOptName in ('-q', '--' + OPT_QUIET):
                if not lLogLevels.isDebug and not lLogLevels.isVerbose:
                    lLogLevels.isQuiet = True
                else:
                    pass #print ERROR, 'quiet option conflicts with \
                    #previous option.'
            if lOptName in ('-t', '--' + OPT_TRANSLATE):
                self.__export = lOptArg
            if lOptName in ('-u', '--' + OPT_USER):
                self.__user = lOptArg.lower()


        return lLogLevels
        
class MainServer(asyncore.dispatcher):
    def __init__(self, aPort, aLogger):
        asyncore.dispatcher.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.bind(('', aPort))
        self.listen(5)

        self.__logger = aLogger
        aLogger.inform('NetFrogd main server initialized on port ' + \
                       str(aPort))

    def handle_accept(self):
        aSock, aAddress = self.accept()
        FrogServer(self.__logger, aSock)


class FrogServer(asynchat.async_chat):
    def __init__(self, aLogger, *aArgs):
        asynchat.async_chat.__init__(self, *aArgs)
        self.set_terminator('\n')

        self.__logger = aLogger
        
        self.__root = ''
        self.__file = None
        self.__fileTimestamp = None

        self.__data = ''
        self.__user = User(self.getpeername())
        self.__logger.inform('Frog individual server initialized for ' + \
                             self.__user())

    def authenticate(self, aUser):
        self.__logger.log(aUser + ' authenticated on ' + self.__user())
        self.__user.name = aUser

        # Prepare the roots
        self.__root = os.path.join(DATA_ROOT, self.__user.name)
        if not os.path.isdir(self.__root):
            os.makedirs(self.__root)
        os.chdir(self.__root)

        self.__logger.inform(self.__user.name + ' data stored under ' + \
                             self.__root)
        self.__checkFile(self.__getWeekTimeTuple())

    def storeData(self, aFrogData):
        """Write the data to a file.

        Format is dd/MM/yyyy:hh/mm:temp:pres:hygro:[comment]
        """
        if self.__user.name == '':
            self.__logger.warn('Data received without authentification.' + \
                               'Please connect first.')
            return

        lTimeTuple = self.__getTimeTuple()
        self.__checkFile(lTimeTuple)
        
        lData, lComment = aFrogData.split(':')
        lTemperature, lPressure, lHygroscopy = lData.split(' ')
        self.__logger.log('%sdeg, %smBar, %s%% from %s' % \
                          (lTemperature, lPressure, lHygroscopy,self.__user()))
        
        lLine = self.__formatDate(lTimeTuple) + ':' \
                + lTemperature + ':' + lPressure + ':' + lHygroscopy \
                + ':' + lComment.replace(' ', '%20') + os.linesep
        
        self.__file.write(lLine)
        self.__file.flush()
        
    def collect_incoming_data(self, aData):
        self.__logger.debug('Received %d bytes from %s ' % \
                            (len(aData), self.__user()))
        self.__data = ''.join(aData)

    def found_terminator(self):
        if self.__data.startswith('user'):
            self.authenticate(self.__data[4:].strip().lower())
        elif self.__data.startswith('data'):
            self.storeData(self.__data[4:].strip())
        else:
            self.__logger.warn('Unknown command from ' + self.__user())
        self.__data = []

    def handle_close(self):
        self.__logger.log(self.__user() + ' disconnected.')
        self.close()
        if self.__file != None:
            self.__file.close()

    def __getTimeTuple(self):
        """Returns the 9-tuple of this very moment.
        """
        return time.localtime(time.time())     

    def __getWeekTimeTuple(self):
        """Returns the 9-tuple of the beginning of this week.
        """
        lNowTuple = self.__getTimeTuple()
        if lNowTuple.tm_wday == 6:
            return lNowTuple
        
        lDelta = datetime.timedelta(days = - (lNowTuple.tm_wday + 2))
        lWeekDate = datetime.date.today() + lDelta
        return lWeekDate.timetuple()     

    def __formatDate(self, aTimeTuple):
        """Returns the date formatted to fit an entry in data files.
        """
        return time.strftime('%d/%m/%Y:%H/%M', aTimeTuple)

    def __formatFilename(self, aTimeTuple):
        """Returns data filename for the week beginning at the given date.
        """
        lFilename = time.strftime('%Y%m%d', aTimeTuple)
        return self.__user.name + '-' + lFilename

    def __checkFile(self, aTimeTuple):
        """Checks if a new file must be opened.

        The data from one user is kept in files under
        DATA_ROOT/user_name. Each file holds the data produced during
        one week (from saturday to friday midnight). The filenames are
        named after the first day of the week they represent (ie
        'machine-20030401'). Old data files are zipped.

        """
        if self.__user.name == '':
            self.__logger.warn('__checkFile: Internal Error:' + \
                               'no user authenticated.')
            return
            
        if self.__file == None:
            lFilename = self.__formatFilename(aTimeTuple)
            self.__file = open(lFilename, 'a')
            self.__fileTimestamp = aTimeTuple
            self.__logger.log('Logging ' + self.__user() + \
                              ' to ' + lFilename)
            return

        if aTimeTuple.tm_wday == 6 and \
           aTimeTuple.tm_mday != self.__fileTimestamp.tm_mday:

             # If this day is a new saturday, we must zip the file...
            self.__file.close()
            lZipFile = zipfile.ZipFile(self.__file.name + '.zip', 'w')
            lZipFile.write(self.__file.name)
            lZipFile.close()
            self.__logger.log('Week data zipped in ' + lZipFile.name)
            
            # ... and begin a new one 
            lFilename = self.__formatFilename(aTimeTuple)
            self.__file = open(lFilename, 'w')
            self._fileTimestamp = aTimeTuple
            self.__logger.log('Now logging ' + self.__user() + \
                              ' to ' + lFilename)


#
# Main
#
if __name__ == "__main__":
    print '## Frog Python Daemon.', VERSION
    lDispatcher = None

    try:
        lDispatcher = Dispatcher()
        lDispatcher.process()
    except KeyboardInterrupt:
        print '## Interrupted by user.'
    except Exception, e:
        print e

    lDispatcher.interrupt()
