# netfrog. A distributed platform to manage with frog data.
# Copyright (C) 2004 David Soulayrol <dsoulayrol@free.fr>
#
# This file is part of netfrog.
#
# netfrog is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
A centralized server for frog data storage.

How can we move the Frog device (http://froggyhome.com) around on many
computers and collect all the data on one server at realtime pace ? 
The software distributed with the Frog doesn't help. It allows FTP
export with nice graphs, but nothing as sophisticated as we want.

NetFrogd is the answer. This daemon waits for connections, and then
stores the data sent by distant clients. Clients must authentify
first, and data is stored per machine.

[TODO: rewrite this}
"""

import asyncore, asynchat, socket
import getopt, imp, os, sys
import getopt

from netfrog.config.common import ConfigurationParser, ServerConfiguration, Machine
import ground.util.log as log


# For introspection
import netfrog.server.services as services


VERSION = 'v0.4'
DATA_ROOT='./test/'


# Textual data
NOTICE = """NetFrogd version 0.4, Copyright (C) 2004 David Soulayrol
NetFrogd comes with ABSOLUTELY NO WARRANTY; for details type `NetFrogd -w'.
This is free software, and you are welcome to redistribute it under
certain conditions; see the COPYING file.
"""

WARRANTY = """  NO WARRANTY

  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
mERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS
TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
REPAIR OR CORRECTION.

  12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,
INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING
OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED
TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY
YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER
PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGES.
"""


class ServiceBroker(object):
    def __init__(self, aUser, aServicesConfigurationList):
        self.__configurations = aServicesConfigurationList
        self.__services = {}
        self.__user = aUser


    def startService(self, aService):
        log.Logger.info('starting %s', aService.getName())
        lFile, lPath, lHelp = imp.find_module(
            aService.getModuleName(), services.__path__)
        lModule = imp.load_module(
            aService.getModuleName(), lFile, lPath, lHelp)
        lClass = lModule.__dict__.get(aService.getName())
        self.__services[aService.getName()] = lClass(
            self.__user, aService.getParameters())


    def stopService(self, aService):
        log.Logger.info('stopping %s', aService.getName())
        # TODO
        

    def startServices(self):
        try:
            for lService in self.__configurations.values():
                self.startService(lService)
        except Exception, e:
            print e


    def stopServices(self):
        try:
            for lService in self.__configurations.values():
                self.stopService(lService)
        except Exception, e:
            print e


    def findServiceFor(self, aMessage):
        lCommand = aMessage.split(' ')[0].strip()
        log.Logger.debug('Searching service for command \'%s\'' % \
                         (lCommand))
        for lService in self.__configurations.values():
            if lCommand in lService.getCommands():
                return self.__services[lService.getName()]
        return None



class ServiceSocketServer(asynchat.async_chat):
    def __init__(self, *aArgs):
        asynchat.async_chat.__init__(self, *aArgs)
        self.set_terminator('\n')

        self.__data = []
        self.__user = None
        self.__configuration = None
        self.__broker = None

    def setup(self, aConfiguration):
        self.__user = Machine.getMachineFromPair(self.getpeername())
        self.__configuration = aConfiguration
        self.__broker = ServiceBroker(self.__user,
                                      aConfiguration.getServices())
        self.authenticate()
        self.__broker.startServices()
        
        
    def authenticate(self):
        if self.__user.name in self.__configuration.getAuthorizedUsers():
            log.Logger.info('authentication succeeded for %s' % (self.__user))
        else:
            log.Logger.info('authentication failed for %s' % (self.__user))
            # TODO: use constant error quotes
            self.push('101 Authentication failed\n')
            self.close()
            

    def route(self, aMessage):
        lService = self.__broker.findServiceFor(aMessage)
        if lService == None:
            log.Logger.error('cannot route \'%s\' message for %s' % \
                             (aMessage, self.__user))
        else:
            lService.process(aMessage)
            
            
    def collect_incoming_data(self, aData):
        log.Logger.debug('Received %d bytes from %s' % \
                         (len(aData), self.__user))
        self.__data.append(aData)


    def found_terminator(self):
        lMessage = ''.join(self.__data).strip()
        log.Logger.debug('Read message \'%s\' from %s' % \
                         (lMessage, self.__user))
        self.route(lMessage)
        self.__data = []


    def handle_close(self):
        log.Logger.info(str(self.__user) + ' disconnected.')
        self.close()



class Server(asyncore.dispatcher):
    def __init__(self):
        asyncore.dispatcher.__init__(self)
        lFilename = self.__readCommandLine()
        self.__configuration = ServerConfiguration(lFilename)
        self.__parseConfigurationFile()


    def run():
        print '## Frog Python Daemon.', VERSION
        try:
            lServer = Server()
            lServer.start()
            asyncore.loop()
        except KeyboardInterrupt, e:
            print e
        except Exception, e:
            print e
    run = staticmethod(run)


    def start(self):
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.bind(('', self.__configuration.getPort()))
        self.listen(5)


    def handle_accept(self):
        """Accept and manage a new connection.
        """
        aSock, aAddress = self.accept()
        ServiceSocketServer(aSock).setup(self.__configuration)

    
    def __parseConfigurationFile(self):
        """Parse a configuration file and return a ServerConfiguration.
        """
        lParser = ConfigurationParser(self.__configuration)
        lParser.parse(self.__configuration.getFilename())

    
    def __printUsage(self):
        """Prints a short notice about NetFrog usage.
        """
        print NOTICE
        print 'Usage: netfrog [-c|--conditions] [-w|--warranty] config'


    def __printGplConditions(self):
        """Prints the GPL license.
        """
        print "[TODO] conditions..."


    def __printGplWarranty(self):
        """Prints the warranty section of the GPL license.
        """
        print WARRANTY
        

    def __readCommandLine(self): 
        """Analyse command line.
        """
        # Command line options
        OPT_CONDITIONS = 'conditions'
        OPT_WARRANTY = 'warranty'
        OPT_DEBUG = 'debug'
        OPT_VERBOSE = 'verbose'
        OPT_QUIET = 'quiet'

        OPT_LIST = 'cwdvq'
        LONG_OPT_LIST = [OPT_CONDITIONS, OPT_WARRANTY,
                         OPT_DEBUG, OPT_VERBOSE, OPT_QUIET]

        lLogLevel = log.WARNING
        
        try:
            lOptlist, lArgs = getopt.getopt(sys.argv[1:],
                                            OPT_LIST,
                                            LONG_OPT_LIST)
        except getopt.GetoptError:
            self.__printUsage()
            sys.exit()

        for lOptName, lOptArg in lOptlist:
            if lOptName in ('-c', '--' + OPT_CONDITIONS):
                self.__printGplConditions()
                sys.exit()
            if lOptName in ('-w', '--' + OPT_WARRANTY):
                self.__printGplWarranty()
                sys.exit()

            if lOptName in ('-d', '--' + OPT_DEBUG):
                if lLogLevel != 0:
                    lLogLevel = log.DEBUG
                else:
                    print ERROR, 'debug option conflicts with quiet.'
            if lOptName in ('-v', '--' + OPT_VERBOSE):
                if lLogLevel != 0:
                    lLogLevel = log.INFO
                else:
                    print ERROR, 'verbose option conflicts with quiet.'
            if lOptName in ('-q', '--' + OPT_QUIET):
                if lLogLevel != log.DEBUG \
                       and lLogLevel != log.VERBOSE:
                    lLogLevel = 0
                else:
                    print ERROR, 'quiet option conflicts with \
                    #previous option.'

        if len(lArgs) != 1:
            raise Exception('No configuration.')

        # Configure all future instances of Logger
        log.Logger('netfrog', lLogLevel)
        log.Logger.info('## Frog Python Daemon %s new session using %s',
                        VERSION, lArgs[0])

        # Return the configuration file
        return lArgs[0]
